@@ -1,5 +1,4 @@
 from typing import Optional, Callable, Iterator, Iterable, List, Set, Mapping, Tuple, Any, Dict
-from functools import cmp_to_key
 import ast
 import warnings
 import importlib
@@ -19,21 +18,54 @@ _visited_objects = []
 function_docstring_preprocessing_hooks: List[Callable[[str], str]] = []
 
 
-class DirectoryWalkerGuard(object):
+def _find_str_end(s, start):
+    for i in range(start, len(s)):
+        if i == "\\":  # skip escaped chars
+            continue
+        if i == s[start]:
+            return i
+    return -1
 
+
+def _is_balanced(s):
+    closing = {
+        "(": ")",
+        "{": "}",
+        "[": "]"
+    }
+
+    stack = []
+    i = 0
+    while i < len(s):
+        c = s[i]
+        if c in "\"'":
+            # TODO: handle triple-quoted strings too
+            i = _find_str_end(s, i)
+            if i < 0:
+                return False
+        if c in closing:
+            stack.append(closing[c])
+        elif stack and stack[-1] == c:
+            stack.pop()
+        i += 1
+
+    return len(stack) == 0
+
+
+class DirectoryWalkerGuard(object):
     def __init__(self, dirname):
         self.dirname = dirname
+        self.origin = os.getcwd()
 
     def __enter__(self):
         if not os.path.exists(self.dirname):
-            os.mkdir(self.dirname)
+            os.makedirs(self.dirname)
 
         assert os.path.isdir(self.dirname)
-
         os.chdir(self.dirname)
 
     def __exit__(self, exc_type, exc_val, exc_tb):
-        os.chdir(os.path.pardir)
+        os.chdir(self.origin)
 
 
 _default_pybind11_repr_re = re.compile(r'(<(?P<class>\w+(\.\w+)*) object at 0x[0-9a-fA-F]+>)|'
@@ -196,7 +228,7 @@ class StubsGenerator(object):
         re.compile(
             r"numpy.ndarray\[(?P<type>[^\[\]]+)(\[(?P<shape>[^\[\]]+)\])?(?P<extra>[^][]*)\]"): replace_numpy_array,
         re.compile(
-            r"(?<!\w)(?P<type>Callable|Dict|[Ii]terator|[Ii]terable|List|Optional|Set|Tuple|Union')(?!\w)"): replace_typing_types
+            r"(?<!\w)(?P<type>Callable|Dict|[Ii]terator|[Ii]terable|List|Optional|Set|Tuple|Union)(?!\w)"): replace_typing_types
     }
 
     def parse(self):
@@ -215,6 +247,14 @@ class StubsGenerator(object):
         lines = [StubsGenerator._indent(l) if l else l for l in lines]
         return "\n".join(lines)
 
+    @staticmethod
+    def is_valid_module(module_name):  # type: (str) -> bool
+        import importlib.util
+        try:
+            return importlib.import_module(module_name) is not None
+        except ModuleNotFoundError:
+            return False
+
     @staticmethod
     def fully_qualified_name(klass):
         module_name = klass.__module__ if hasattr(klass, '__module__') else None
@@ -234,17 +274,13 @@ class StubsGenerator(object):
         return s
 
     @staticmethod
-    def function_signatures_from_docstring(name, func, module_name):  # type: (Any, str) -> List[FunctionSignature]
+    def function_signatures_from_docstring(name, func, module_name):  # type: (str, Any, str) -> List[FunctionSignature]
         try:
-            no_parentheses = r"[^()]*"
-            parentheses_one_fold = r"({nopar}(\({nopar}\))?)*".format(nopar=no_parentheses)
-            parentheses_two_fold = r"({nopar}(\({par1}\))?)*".format(par1=parentheses_one_fold, nopar=no_parentheses)
-            parentheses_three_fold = r"({nopar}(\({par2}\))?)*".format(par2=parentheses_two_fold, nopar=no_parentheses)
             signature_regex = r"(\s*(?P<overload_number>\d+).)" \
                               r"?\s*{name}\s*\((?P<args>{balanced_parentheses})\)" \
                               r"\s*->\s*" \
                               r"(?P<rtype>[^\(\)]+)\s*".format(name=name,
-                                                               balanced_parentheses=parentheses_three_fold)
+                                                               balanced_parentheses=".*")
             docstring = func.__doc__
 
             for hook in function_docstring_preprocessing_hooks:
@@ -256,7 +292,8 @@ class StubsGenerator(object):
                 if m:
                     args = m.group("args")
                     rtype = m.group("rtype")
-                    signatures.append(FunctionSignature(name, args, rtype))
+                    if _is_balanced(args):
+                        signatures.append(FunctionSignature(name, args, rtype))
 
             # strip module name if provided
             if module_name:
@@ -339,6 +376,7 @@ class StubsGenerator(object):
 
     @staticmethod
     def format_docstring(docstring):
+        docstring = inspect.cleandoc("\n" + docstring)
         return StubsGenerator.indent('"""\n{}\n"""'.format(docstring.strip("\n")))
 
 
@@ -460,8 +498,9 @@ class FreeFunctionStubsGenerator(StubsGenerator):
         for s in self.signatures:  # type: FunctionSignature
             for t in s.get_all_involved_types():  # type: str
                 try:
-                    i = t.rindex(".")
-                    involved_modules_names.add(t[:i])
+                    module_name = t[:t.rindex(".")]
+                    if self.is_valid_module(module_name):
+                        involved_modules_names.add(module_name)
                 except ValueError:
                     pass
         return involved_modules_names
@@ -753,6 +792,11 @@ class ModuleStubsGenerator(StubsGenerator):
         if self.doc_string:
             result += ['"""' + self.doc_string.replace('"""', r'\"\"\"') + '"""']
 
+        if sys.version_info[:2] >= (3, 7):
+            result += [
+                "from __future__ import annotations"
+            ]
+
         result += [
             "import {}".format(self.module.__name__)
         ]
@@ -804,10 +848,6 @@ class ModuleStubsGenerator(StubsGenerator):
         return self.module.__name__.split(".")[-1]
 
     def write(self):
-        if not os.path.exists(self.short_name + self.stub_suffix):
-            logger.debug("mkdir `%s`" % (self.short_name + self.stub_suffix))
-            os.mkdir(self.short_name + self.stub_suffix)
-
         with DirectoryWalkerGuard(self.short_name + self.stub_suffix):
             with open("__init__.pyi", "w") as init_pyi:
                 init_pyi.write("\n".join(self.to_lines()))
@@ -842,16 +882,6 @@ setup(
 )""".format(package_name=self.short_name))
 
 
-def recursive_mkdir_walker(subdirs, callback):  # type: (List[str], Callable) -> None
-    if len(subdirs) == 0:
-        callback()
-    else:
-        if not os.path.exists(subdirs[0]):
-            os.mkdir(subdirs[0])
-        with DirectoryWalkerGuard(subdirs[0]):
-            recursive_mkdir_walker(subdirs[1:], callback)
-
-
 def main(args=None):
     parser = ArgumentParser(prog='pybind11-stubgen', description="Generates stubs for specified modules")
     parser.add_argument("-o", "--output-dir", help="the root directory for output stubs", default="./stubs")
@@ -905,19 +935,17 @@ def main(args=None):
         handlers=handlers
     )
 
-    output_path = sys_args.output_dir
-
-    if not os.path.exists(output_path):
-        os.mkdir(output_path)
-
-    with DirectoryWalkerGuard(output_path):
+    with DirectoryWalkerGuard(sys_args.output_dir):
         for _module_name in sys_args.module_names:
             _module = ModuleStubsGenerator(_module_name)
             _module.parse()
             if FunctionSignature.n_fatal_errors() == 0:
                 _module.stub_suffix = sys_args.root_module_suffix
                 _module.write_setup_py = not sys_args.no_setup_py
-                recursive_mkdir_walker(_module_name.split(".")[:-1], lambda: _module.write())
+                _dir = _module_name.split(".")[:-1] or ["."]
+                _dir = os.path.join(*_dir)
+                with DirectoryWalkerGuard(_dir):
+                    _module.write()
 
         if FunctionSignature.n_invalid_signatures > 0:
             logger.info("Useful link: Avoiding C++ types in docstrings:")
